// File dstring.h for CC1B
#ifndef _DSTRING_H_
#define _DSTRING_H_
//include other libraries here
#include <system/memory.h>
#include <text/string/string.h>
#pragma library 1

/**
 * \file dstring.h
 * \brief Library for constant string support
 *
 * This library provides functions for strings located in rom.\n
 * Unlike library cstring.h that uses const char * pointers, this\n
 * library uses long for pointers.
 * \n
 */
 
/**
 * \brief Calculate the length of a constant string
 *
 * The dstrlen() function calculates the length of the constant string s, not including the terminating `\0' character.
 *
 * \n
 * \param s Address of string in rom.
 * \return The dstrlen() function returns the length of constant string s.
 * \n
 */
long dstrlen(long s)
{
  long n = -1;
  char temp;
  do {
    RomWord(s);
    temp = W;
    s++;
    n++;
  } while (temp);
  return (n);
}

/**
 * \brief Concatenate constant string to string
 *
 * The dstrcat() function appends the t string to the s string overwriting the `\0' character\n
 * at the end of s, and then adds a terminating `\0' character.  The strings may not overlap,\n
 * and the s string must have enough space for the result.
 *
 * \n
 * \param s Address of string in ram.
 * \param t Address of string in rom.
 * \return The dstrcat() function returns a pointer to the destination string s.
 * \n
 */
char *dstrcat(char *s, long t)
{
  char *d = s + strlen(s);
  char temp;
  do {
    RomWord(t);
    temp = W;
	*d = temp;
	d++;
	t++;
  } while (!temp);
  return (s);
}

/**
 * \brief Locate character in constant string
 *
 * The dstrchr() function searches a constant string for the presence of a specific character.
 *
 * \n
 * \param s Address of string in rom.
 * \param c Character to locate.
 * \return The dstrchr() function returns a pointer to the matched character or NULL if the character is not found.
 * \n
 */
long dstrchr(long s, char c)
{
  char temp;
  do {
    RomWord(s);
    temp = W;
    if (temp == c) return (s);
    ++s;
  } while (!temp);
  return (0);
}

/**
 * \brief Compare a string against a constant string
 *
 * The dstrcmp() function compares the two strings s and t.\n
 * It returns an integer less than, equal to, or greater than zero if s is found,\n
 * respectively, to be less than, to match, or be greater than t.
 *
 * \n
 * \param s Address of string in ram.
 * \param t Address of string in rom.
 * \return The dstrcmp() function returns an integer less than, equal to, or greater than zero\n
 * if s is found, respectively, to be less than, to match, or be greater than t.
 * \n
 */
int dstrcmp(char *s, long t)
{
  char temp;
  do {
    RomWord(t);
    temp = W;
    if (*s != temp) break;
    if (!*s) return (0);
    ++s;
    ++t;
  } while (1); 
  return (*s - temp);
}

/**
 * \brief Copy constant string to string
 *
 * The dstrcpy() function copies the string pointed to by t (including the terminating `\0' character)\n
 * to the array pointed to by s. The strings may not overlap, and the destination string s must be\n
 * large enough to receive the copy.
 *
 * \n
 * \param s Address of string in ram.
 * \param t Address of string in rom.
 * \return The dstrcpy() function returns a pointer to the destination string s.
 * \n
 */
char *dstrcpy(char *s, long t)
{
  char *d = s;
  char temp;
  do {
    RomWord(t);
    temp = W;
	*d = temp;
	d++;
	t++;
  } while (temp);
  return (s);
}

/**
 * \brief Concatenate constant string to string for specified number of characters
 *
 * The dstrncat() function appends up to the specified number of characters from the t string to the\n
 * s string overwriting the `\0' character at the end of s, and then adds a terminating `\0' character.\n
 * The strings may not overlap, and the s string must have enough space for the result.
 *
 * \n
 * \param s Address of string in ram.
 * \param t Address of string in rom.
 * \param len Number of characters to copy.
 * \return The dstrncat() function returns a pointer to the destination string s.
 * \n
 */
char *dstrncat(char *s, long t, int len)
{
  char *d = s + strlen(s);
  char temp;
  while (len) {
    len--;
	RomWord(t);
	temp = W;
    if (!temp) break;
    *d = temp;
	d++;
	t++;
  }
  *d = 0;
  return (s);
}

/**
 * \brief Compare a string against a constant string for up to a specified number of bytes
 *
 * The dstrncmp() function compares the two strings s and t for the first (at most) n characters.\n
 * It returns an integer less than, equal to, or greater than zero if s is (or the first n bytes thereof)\n
 * found, respectively, to be less than, to match, or be greater than t.
 *
 * \n
 * \param s Address of string in ram.
 * \param t Address of string in rom.
 * \param len Number of characters to compare.
 * \return The dstrncmp() function returns an integer less than, equal to, or greater than zero\n
 * if s (or the first n bytes thereof) is found, respectively, to be less than, to match, or be greater than t.
 * \n
 */
int dstrncmp(char *s, long t, int len)
{
  char temp;
  while (len) {
    RomWord(t);
    temp = W;
    if (*s != temp) break;
    if (!*s) return (0);
    ++s;
    ++t;
    --len;
  }
  if (len) return (*s - temp);
  return (0);
}

/**
 * \brief Copy constant string to string for specified number of characters
 *
 * The dstrncpy() function copies up to the specified number of characters from the t string\n
 * (including the terminating `\0' character) to the array pointed to by s. The strings may not overlap,\n
 * and the destination string s must be large enough to receive the copy.
 *
 * \n
 * \param s Address of string in ram.
 * \param t Address of string in rom.
 * \param len Number of bytes to copy.
 * \return The dstrncpy() function returns a pointer to the destination string s.
 * \n
 */
char *dstrncpy(char *s, long t, int len)
{
  char *d = s;
  char temp;
  while (len) {
    RomWord(t);
    temp = W;
    *d = temp;
	if (!temp) break;
	d++;
	t++;
    len--;
	continue;
    while (len) {
	  *d++ = 0;
	  len--;
    }
  }
  return (s);
}

/**
 * \brief Locate character in constant string
 *
 * The dstrrchr() function searches a constant string for the presence of a specific character.
 *
 * \n
 * \param s Address of string in rom.
 * \param c Character to locate.
 * \return The dstrrchr() function returns a pointer to the last occurrence of the character c in the string s.
 * \n
 */
long dstrrchr(long s, char c)
{
  const char *ptr = 0;
  char temp;
  do {
    RomWord(s);
    temp = W;
    if (temp == c) ptr = s;
    ++s;
  } while (temp);
  return (ptr);
}

#pragma library 0
#endif
// End of file dstring.h
