// File vpUart.h for CC1B
#ifndef _VPUART_H_
#define _VPUART_H_
//include other libraries here
#include <system/memory.h>
#pragma library 1

/**
 * \file vpUart.h
 * \brief Library for virtual peripheral vpUart
 *
 * This library provides virtual peripheral uarts for both transmit and receive. These uarts\n
 * are fully featured: 7/8 databits, none/even/odd parity, invert/normal mode.
 * \n
 */

#define Baud_divide(baud,ratio) ((2*uartfs/baud/ratio)+1)/2 //divisor for baud timing

/**
 * \brief Define a receive uart
 *
 * The macro vpUartRx() defines a receive uart. Its parameters are stored in rom.\n
 * You install this uart by calling vph = vpInstall(name) and you remove the uart by\n
 * calling vpUninstall(vph).
 *
 * \n
 * How to use:
 * \code
 * vpUartRx(RX1,PORTPIN_RA2+PORTPIN_INV,LOWINPUT_LOWINPUT,0,0,9600,1,E72)
 * char rx1 = vpInstall(RX1);
 * \endcode
 *
 * \n
 * The above defines a receive uart with inverted mode for datapin RA.2 that uses 9600 baud\n
 * for 7 databits with even parity and 2 stopbits. Note the absent ';' after the definition.
 *
 * \n
 * \param name Name for this uart. The name must be unique as it is used as a label for rom data.\n
 *
 * \param datapin Pin to receive data. Use PORTPIN_XX, optionally ORed with PORTPIN_INV\n
 * (inverted mode) and/or PORTPIN_OC (open collector mode).\n
 *
 * \param dataconfig Configuration for the datapin. This specifies the pin states after installation\n
 * and removal of the uart. Its format is XX_YY in which XX defines the state after installation\n
 * and YY defines the state after removal of the uart. Possible values for XX and YY are:\n
 * LEAVE, LOWOUTPUT, HIGHOUTPUT, LOWINPUT, HIGHINPUT. Because the datapin\n
 * is an input pin, only LOWINPUT and HIGHINPUT make sense.\n
 *
 * \param hspin Pin for handshake. Use PORTPIN_XX, optionally ORed with PORTPIN_INV\n
 * (inverted mode) and/or PORTPIN_OC (open collector mode). Use 0 for no handshake.\n
 *
 * \param hsconfig Configuration for the handshake pin. This specifies the pin states after installation\n
 * and removal of the uart. Its format is XX_YY in which XX defines the state after installation\n
 * and YY defines the state after removal of the uart. Possible values for XX and YY are:\n
 * LEAVE, LOWOUTPUT, HIGHOUTPUT, LOWINPUT, HIGHINPUT. Because the handshake pin\n
 * is an output pin, only LOWOUTPUT and HIGHOUTPUT make sense. Use 0 for no handshake.\n
 *
 * \param baud The baudrate for the uart. You can specify the usual baudrates: 1200, 2400, 4800, 9600,\n
 * 19200, 28800, 38400, 57600, 115200. Which baudrates are supported is determined by the\n
 * CPUFREQ and MAXBAUD settings.\n
 *
 * \param format The format for the uart. This specifies the dataformat. Values available are:\n
 * - E71	even parity, 7 databits, 1 stopbit\n
 * - E72	even parity, 7 databits, 2 stopbits\n
 * - E81	even parity, 8 databits, 1 stopbit\n
 * - O71	odd  parity, 7 databits, 1 stopbit\n
 * - O72	odd  parity, 7 databits, 2 stopbits\n
 * - O81	odd  parity, 8 databits, 1 stopbit\n
 * - N71	no   parity, 7 databits, 1 stopbit\n
 * - N72	no   parity, 7 databits, 2 stopbits\n
 * - N81	no   parity, 8 databits, 1 stopbit\n
 * - N82	no   parity, 8 databits, 2 stopbits
 *
 * \n
 */
#define vpUartRx(name,datapin,dataconfig,hspin,hsconfig,baud,format) \
#ifndef vpUartRx_installed \
#define vpUartRx_installed \
#endif \
#if (Baud_divide(baud,VPRATIO) < 2) || (Baud_divide(baud,VPRATIO) > 170) \
#error Specified baudrate not possible with current ratio \
#endif \
#pragma cdata. ##name = VP_UARTRX,16 \
#if (format & 0x18) != 0 \
#ifndef vpUartRx_parity \
#define vpUartRx_parity \
#endif \
#endif \
#if (hspin != 0) \
#ifndef vpUartRx_hs \
#define vpUartRx_hs \
#endif \
#if (hspin & PORTPIN_OC) != 0 \
#ifndef vpUartRx_hoc \
#define vpUartRx_hoc \
#endif \
#endif \
#pragma cdata[] = ((datapin)>>8)|0x80 \
#else \
#pragma cdata[] = (datapin)>>8 \
#endif \
#pragma cdata[] = (datapin)&0xFF,Baud_divide(baud,VPRATIO),format,0,0,0,0 \
#if (hspin != 0) \
#pragma cdata[] = (hspin)>>8,(hspin)&0xFF, 0x40 \
#else \
#pragma cdata[] = 0x60,0,0 \
#endif \
#pragma cdata[] = 0,0,0,0,0 \
#pragma cdata. ##name##_pinconfig \
#if (datapin != 0) && (dataconfig != 0) \
#pragma cdata[] = (datapin)>>8,(datapin)&0xFF,dataconfig \
#endif \
#if (hspin != 0) && (hsconfig != 0) \
#pragma cdata[] = (hspin)>>8,(hspin)&0xFF,hsconfig \
#endif \
#pragma cdata[] = 0

/**
 * \brief Define a transmit uart
 *
 * The macro vpUartTx() defines a transmit uart. Its parameters are stored in rom.\n
 * You install this uart by calling vph = vpInstall(name) and you remove the uart by\n
 * calling vpUninstall(vph).
 *
 * \n
 * How to use:
 * \code
 * vpUartTx(TX1,PORTPIN_RA1,HIGHOUTPUT_HIGHOUTPUT,0,0,9600,1,E72)
 * char tx1 = vpInstall(TX1);
 * \endcode
 *
 * \n
 * The above defines a transmit uart with normal mode for datapin RA.1 that uses 9600 baud\n
 * for 7 databits with even parity and 2 stopbits. Note the absent ';' after the definition.
 *
 * \n
 * \param name Name for this uart. The name must be unique as it is used as a label for rom data.\n
 *
 * \param datapin Pin to transmit data. Use PORTPIN_XX, optionally ORed with PORTPIN_INV\n
 * (inverted mode) and/or PORTPIN_OC (open collector mode).\n
 *
 * \param dataconfig Configuration for the datapin. This specifies the pin states after installation\n
 * and removal of the uart. Its format is XX_YY in which XX defines the state after installation\n
 * and YY defines the state after removal of the uart. Possible values for XX and YY are:\n
 * LEAVE, LOWOUTPUT, HIGHOUTPUT, LOWINPUT, HIGHINPUT. Because the datapin\n
 * is an output pin, only LOWOUTPUT and HIGHOUTPUT make sense.\n
 *
 * \param hspin Pin for handshake. Use PORTPIN_XX, optionally ORed with PORTPIN_INV\n
 * (inverted mode) and/or PORTPIN_OC (open collector mode). Use 0 for no handshake.\n
 *
 * \param hsconfig Configuration for the handshake pin. This specifies the pin states after installation\n
 * and removal of the uart. Its format is XX_YY in which XX defines the state after installation\n
 * and YY defines the state after removal of the uart. Possible values for XX and YY are:\n
 * LEAVE, LOWOUTPUT, HIGHOUTPUT, LOWINPUT, HIGHINPUT. Because the handshake pin\n
 * is an input pin, only LOWINPUT and HIGHINPUT make sense. Use 0 for no handshake.\n
 *
 * \param baud The baudrate for the uart. You can specify the usual baudrates: 1200, 2400, 4800, 9600,\n
 * 19200, 28800, 38400, 57600, 115200. Which baudrates are supported is determined by the\n
 * CPUFREQ and MAXBAUD settings.\n
 *
 * \param format The format for the uart. This specifies the dataformat. Values available are:\n
 * - E71	even parity, 7 databits, 1 stopbit\n
 * - E72	even parity, 7 databits, 2 stopbits\n
 * - E81	even parity, 8 databits, 1 stopbit\n
 * - O71	odd  parity, 7 databits, 1 stopbit\n
 * - O72	odd  parity, 7 databits, 2 stopbits\n
 * - O81	odd  parity, 8 databits, 1 stopbit\n
 * - N71	no   parity, 7 databits, 1 stopbit\n
 * - N72	no   parity, 7 databits, 2 stopbits\n
 * - N81	no   parity, 8 databits, 1 stopbit\n
 * - N82	no   parity, 8 databits, 2 stopbits
 *
 * \n
 */
#define vpUartTx(name,datapin,dataconfig,hspin,hsconfig,baud,format) \
#ifndef vpUartTx_installed \
#define vpUartTx_installed \
#endif \
#if (Baud_divide(baud,VPRATIO) < 2) || (Baud_divide(baud,VPRATIO) > 170) \
#error Specified baudrate not possible with current ratio \
#endif \
#if (datapin & PORTPIN_OC) != 0 \
#ifndef vpUartTx_doc \
#define vpUartTx_doc \
#endif \
#endif \
#pragma cdata. ##name = VP_UARTTX,16 \
#if (format & 0x18) != 0 \
#ifndef vpUartTx_parity \
#define vpUartTx_parity \
#endif \
#endif \
#if (hspin != 0) \
#ifndef vpUartTx_hs \
#define vpUartTx_hs \
#endif \
#pragma cdata[] = ((datapin)>>8)|0x80 \
#else \
#pragma cdata[] = (datapin)>>8 \
#endif \
#pragma cdata[] = (datapin)&0xFF,Baud_divide(baud,VPRATIO),format,0,0,0,0 \
#if (hspin != 0) \
#pragma cdata[] = (hspin)>>8,(hspin)&0xFF, 0x40 \
#else \
#pragma cdata[] = 0x60,0,0 \
#endif \
#pragma cdata[] = 0,0,0,0,0 \
#pragma cdata. ##name##_pinconfig \
#if (datapin != 0) && (dataconfig != 0) \
#pragma cdata[] = (datapin)>>8,(datapin)&0xFF,dataconfig \
#endif \
#if (hspin != 0) && (hsconfig != 0) \
#pragma cdata[] = (hspin)>>8,(hspin)&0xFF,hsconfig \
#endif \
#pragma cdata[] = 0

static shadowDef char vpUart_dport @ 0x10; //data port - lownib=5(RA) to 9(RE)
  static shadowDef bit vpUart_dinvert @ 0x10.4; //invert
  static shadowDef bit vpUart_docmode @ 0x10.5; //opencollector
  static shadowDef bit vpUart_dinput @ 0x10.6; //datapin input placeholder
  static shadowDef bit vpUart_handshake @ 0x10.7; //set if handshake pin used
static shadowDef char vpUart_dpin @ 0x11; //pin bitmask (specified bit set , other bits cleared)
static shadowDef char vpUart_baud @ 0x12; //baudrate
static shadowDef char vpUart_format @ 0x13; //
  static shadowDef bit vpUart_bits0 @ 0x13.0; //total bits: 01=9 10=10 11=11
  static shadowDef bit vpUart_bits1 @ 0x13.1; //
  static shadowDef bit vpUart_7bits @ 0x13.2; //set for 7 databits (0 = 8 databits)
  static shadowDef bit vpUart_even @ 0x13.3; //set for even parity
  static shadowDef bit vpUart_odd @ 0x13.4; //set for odd parity
  static shadowDef bit vpUart_data @ 0x13.5; //set if valid value, cleared after value grabbed
  static shadowDef bit vpUart_error @ 0x13.6; //set if parity error
  static shadowDef bit vpUart_parity @ 0x13.7; //parity bit placeholder
static shadowDef char vpUart_count @ 0x14; //number of bits
static shadowDef char vpUart_divide @ 0x15; //timing counter
static shadowDef char vpUart_low @ 0x16; //shift register
static shadowDef char vpUart_high @ 0x17; //value
static shadowDef char vpUart_hport @ 0x18; //handshake port - lownib=5(RA) to 9(RE) (optional queue if handshake not used)
  static shadowDef bit vpUart_hinvert @ 0x18.4; //invert
  static shadowDef bit vpUart_hocmode @ 0x18.5; //opencollector
  static shadowDef bit vpUart_hinput @ 0x18.6; //handshake pin input placeholder
  static shadowDef bit vpUart_hlevel @ 0x18.7; //logical handshake pin output level
static shadowDef char vpUart_hpin @ 0x19; //pin bitmask (specified bit set , other bits cleared)
static shadowDef char vpUart_buf[6] @ 0x1A; //optional queue

/**
 * \brief Interrupt code for virtual peripheral receive uart
 *
 * This part only receives the serial bits.
 * \n
 */
#define vpUartRx_ISR \
{ \
#pragma update_FSR 0 \
  W = vpUart_dport & 0x0F; \
  FSR = W; \
  _IsrTemp = INDF; \
  FSR = _IsrBank; \
  W = _IsrTemp; \
  if (vpUart_dinvert) W ^= vpUart_dpin; \
  W &= vpUart_dpin; \
  Carry = 1; \
  if (Zero_) Carry = 0; \
  vpUart_dinput = Carry; \
  if (!vpUart_count) { \
    if (!Carry) { \
	  W = vpUart_format; \
	  W &= 0x03; \
	  W |= 0x08; \
      vpUart_count = W; \
    } \
	W = vpUart_baud >> 1; \
    vpUart_divide = W; \
  } \
  vpUart_divide--; \
  if (!vpUart_divide) { \
    vpUart_count--; \
    if (vpUart_count > 0) { \
	  Carry = vpUart_dinput; \
      vpUart_low = rr(vpUart_low); \
      vpUart_parity = Carry; \
    } \
    else { \
      W = vpUart_low; \
      vpUart_high = W; \
      vpUart_data = 1; \
    } \
#ifdef vpUartRx_hs \
    if (vpUart_handshake) { \
      if (vpUart_hlevel) { \
        W = ~vpUart_hpin; \
	    _IsrTemp = W; \
	    W = vpUart_hport & 0x0F; \
	    if (vpUart_hocmode) W |= 0xF0; \
	    FSR = W; \
	    W = _IsrTemp; \
	    INDF &= W; \
      } \
	  else { \
        W = vpUart_hpin; \
	    _IsrTemp = W; \
	    W = vpUart_hport & 0x0F; \
#ifdef vpUartRx_hoc \
	    if (vpUart_hocmode) W |= 0xF0; \
#endif \
	    FSR = W; \
	    W = _IsrTemp; \
	    INDF |= W; \
	  } \
      W = _IsrBank; \
      FSR = W; \
#ifdef vpUartRx_hoc \
      if (vpUart_hocmode) { \
	    vpUart_divide = MODE; \
	    W = vpUart_hport | 0xF0; \
	    FSR = W; \
#if defined _CHIP_SX18_ || defined _CHIP_SX20_ || defined _CHIP_SX28_ \
		W = 0x0F; \
#endif \
#if defined _CHIP_SX48_ || defined _CHIP_SX52_ \
		W = 0x1F; \
#endif \
        MODE = W; \
	    W = FSR & 0x0F; \
	    switch (W) { \
          case PORT_RA: W = DDRACOPY; \
		                controlA(W); \
	                    break; \
          case PORT_RB: W = DDRBCOPY; \
		                controlB(W); \
		                break; \
#if defined _CHIP_SX28_ || defined _CHIP_SX48_ || defined _CHIP_SX52_ \
		  case PORT_RC: W = DDRCCOPY; \
		                controlC(W); \
		                break; \
#endif \
#if defined _CHIP_SX48_ || defined _CHIP_SX52_ \
          case PORT_RD: W = DDRDCOPY; \
		                controlD(W); \
		                break; \
          case PORT_RE: W = DDRECOPY; \
		                controlE(W); \
		                break; \
#endif \
	    } \
        W = _IsrBank; \
        FSR = W; \
	    MODE = vpUart_divide; \
	  } \
#endif \
    } \
#endif \
    vpUart_divide = vpUart_baud; \
  } \
#pragma update_FSR 1 \
}

/**
 * \brief Interrupt code for virtual peripheral receive uart
 *
 * Identical to the macro vpUartRx_ISR.
 * This part only receives the serial bits.
 * \n
 */
void vpUartRx_isr()
{
#pragma update_FSR 0
  FSR = _IsrBank;
  W = vpUart_dport & 0x0F;
  FSR = W;
  _IsrTemp = INDF;
  FSR = _IsrBank;
  W = _IsrTemp;
  if (vpUart_dinvert) W ^= vpUart_dpin;
  W &= vpUart_dpin;
  Carry = 1;
  if (Zero_) Carry = 0;
  vpUart_dinput = Carry;
  if (!vpUart_count) {
    if (!Carry) {
	  W = vpUart_format;
	  W &= 0x03;
	  W |= 0x08;
      vpUart_count = W;
    } 
	W = vpUart_baud >> 1;
    vpUart_divide = W;
  }
  vpUart_divide--;
  if (!vpUart_divide) {
    vpUart_count--;
    if (vpUart_count > 0) {
	  Carry = vpUart_dinput;
      vpUart_low = rr(vpUart_low);
      vpUart_parity = Carry;
    }
    else {
      W = vpUart_low;
      vpUart_high = W;
      vpUart_data = 1;
    }
#ifdef vpUartRx_hs
    if (vpUart_handshake) {
      if (vpUart_hlevel) {
        W = ~vpUart_hpin;
	    _IsrTemp = W;
	    W = vpUart_hport & 0x0F;
	    if (vpUart_hocmode) W |= 0xF0;
	    FSR = W;
	    W = _IsrTemp;
	    INDF &= W;
      }
	  else {
        W = vpUart_hpin;
	    _IsrTemp = W;
	    W = vpUart_hport & 0x0F;
#ifdef vpUartRx_hoc
	    if (vpUart_hocmode) W |= 0xF0;
#endif
	    FSR = W;
	    W = _IsrTemp;
	    INDF |= W;
	  }
      W = _IsrBank;
      FSR = W;
#ifdef vpUartRx_hoc
      if (vpUart_hocmode) {
	    vpUart_divide = MODE;
	    W = vpUart_hport | 0xF0;
	    FSR = W;
#if defined _CHIP_SX18_ || defined _CHIP_SX20_ || defined _CHIP_SX28_
		W = 0x0F;
#endif
#if defined _CHIP_SX48_ || defined _CHIP_SX52_
		W = 0x1F;
#endif
        MODE = W;
	    W = FSR & 0x0F;
	    switch (W) {
          case PORT_RA: W = DDRACOPY;
		                controlA(W);
	                    break;
          case PORT_RB: W = DDRBCOPY;
		                controlB(W);
		                break;
#if defined _CHIP_SX28_ || defined _CHIP_SX48_ || defined _CHIP_SX52_
		  case PORT_RC: W = DDRCCOPY;
		                controlC(W);
		                break;
#endif
#if defined _CHIP_SX48_ || defined _CHIP_SX52_
          case PORT_RD: W = DDRDCOPY;
		                controlD(W);
		                break;
          case PORT_RE: W = DDRECOPY;
		                controlE(W);
		                break;
#endif
	    }
        W = _IsrBank;
        FSR = W;
	    MODE = vpUart_divide;
	  }
#endif
    }
#endif
    vpUart_divide = vpUart_baud;
  }
#pragma update_FSR 1
}

/**
 * \brief Interrupt code for virtual peripheral transmit uart
 *
 * This part only transmits the serial bits.
 * \n
 */
#define vpUartTx_ISR \
{ \
#pragma update_FSR 0 \
  vpUart_divide--; \
  if (!vpUart_divide) { \
    if (vpUart_count) { \
      Carry = 0; \
	  if (vpUart_dinvert) Carry = 1; \
      vpUart_high = rr(vpUart_high); \
      vpUart_low = rr(vpUart_low); \
      vpUart_count--; \
	  if (!vpUart_count) vpUart_data = 0; \
      if (vpUart_low.5) { \
        W = ~vpUart_dpin; \
	    _IsrTemp = W; \
	    W = vpUart_dport & 0x0F; \
#ifdef vpUartTx_doc \
	    if (vpUart_docmode) W |= 0xF0; \
#endif \
	    FSR = W; \
		W = _IsrTemp; \
		INDF &= W; \
      } \
	  else { \
        W = vpUart_dpin; \
	    _IsrTemp = W; \
	    W = vpUart_dport & 0x0F; \
#ifdef vpUartTx_doc \
	    if (vpUart_docmode) W |= 0xF0; \
#endif \
	    FSR = W; \
		W = _IsrTemp; \
		INDF |= W; \
	  } \
      W = _IsrBank; \
      FSR = W; \
#ifdef vpUartTx_doc \
      if (vpUart_docmode) { \
		vpUart_divide = MODE; \
	    W = vpUart_dport | 0xF0; \
		FSR = W; \
#if defined _CHIP_SX18_ || defined _CHIP_SX20_ || defined _CHIP_SX28_ \
		W = 0x0F; \
#endif \
#if defined _CHIP_SX48_ || defined _CHIP_SX52_ \
		W = 0x1F; \
#endif \
        MODE = W; \
		W = FSR & 0x0F; \
		switch (W) { \
          case PORT_RA: W = DDRACOPY; \
		                controlA(W); \
		                break; \
          case PORT_RB: W = DDRBCOPY; \
		                controlB(W); \
		                break; \
#if defined _CHIP_SX28_ || defined _CHIP_SX48_ || defined _CHIP_SX52_ \
		  case PORT_RC: W = DDRCCOPY; \
		                controlC(W); \
		                break; \
#endif \
#if defined _CHIP_SX48_ || defined _CHIP_SX52_ \
          case PORT_RD: W = DDRDCOPY; \
		                controlD(W); \
		                break; \
          case PORT_RE: W = DDRECOPY; \
		                controlE(W); \
		                break; \
#endif \
		} \
        W = _IsrBank; \
        FSR = W; \
		MODE = vpUart_divide; \
	  } \
#endif \
    } \
    vpUart_divide = vpUart_baud; \
#ifdef vpUartTx_hs \
    if (vpUart_handshake) { \
      W = vpUart_hport & 0x0F; \
      FSR = W; \
      _IsrTemp = INDF; \
      FSR = _IsrBank; \
      W = _IsrTemp; \
      if (vpUart_hinvert) W ^= vpUart_hpin; \
      W &= vpUart_hpin; \
      vpUart_hinput = 1; \
      if (Zero_) vpUart_hinput = 0; \
    } \
#endif \
  } \
#pragma update_FSR 1 \
}

/**
 * \brief Interrupt code for virtual peripheral transmit uart
 *
 * Identical to the macro vpUartTx_ISR.
 * This part only transmits the serial bits.
 * \n
 */
void vpUartTx_isr(void)
{
#pragma update_FSR 0
  FSR = _IsrBank;
  vpUart_divide--;
  if (!vpUart_divide) {
    if (vpUart_count) {
      Carry = 0;
	  if (vpUart_dinvert) Carry = 1;
      vpUart_high = rr(vpUart_high);
      vpUart_low = rr(vpUart_low);
      vpUart_count--;
	  if (!vpUart_count) vpUart_data = 0;
      if (vpUart_low.5) {
        W = ~vpUart_dpin;
	    _IsrTemp = W;
	    W = vpUart_dport & 0x0F;
#ifdef vpUartTx_doc
	    if (vpUart_docmode) W |= 0xF0;
#endif
	    FSR = W;
		W = _IsrTemp;
		INDF &= W;
      }
	  else {
        W = vpUart_dpin;
	    _IsrTemp = W;
	    W = vpUart_dport & 0x0F;
#ifdef vpUartTx_doc
	    if (vpUart_docmode) W |= 0xF0;
#endif
	    FSR = W;
		W = _IsrTemp;
		INDF |= W;
	  }
      W = _IsrBank;
      FSR = W;
#ifdef vpUartTx_doc
      if (vpUart_docmode) {
		vpUart_divide = MODE;
	    W = vpUart_dport | 0xF0;
		FSR = W;
#if defined _CHIP_SX18_ || defined _CHIP_SX20_ || defined _CHIP_SX28_
		W = 0x0F;
#endif
#if defined _CHIP_SX48_ || defined _CHIP_SX52_
		W = 0x1F;
#endif
        MODE = W;
		W = FSR & 0x0F;
		switch (W) {
          case PORT_RA: W = DDRACOPY;
		                controlA(W);
		                break;
          case PORT_RB: W = DDRBCOPY;
		                controlB(W);
		                break;
#if defined _CHIP_SX28_ || defined _CHIP_SX48_ || defined _CHIP_SX52_
		  case PORT_RC: W = DDRCCOPY;
		                controlC(W);
		                break;
#endif
#if defined _CHIP_SX48_ || defined _CHIP_SX52_
          case PORT_RD: W = DDRDCOPY;
		                controlD(W);
		                break;
          case PORT_RE: W = DDRECOPY;
		                controlE(W);
		                break;
#endif
		}
        W = _IsrBank;
        FSR = W;
		MODE = vpUart_divide;
	  }
#endif
    }
    vpUart_divide = vpUart_baud;
#ifdef vpUartTx_hs
    if (vpUart_handshake) {
      W = vpUart_hport & 0x0F;
      FSR = W;
      _IsrTemp = INDF;
      FSR = _IsrBank;
      W = _IsrTemp;
      if (vpUart_hinvert) W ^= vpUart_hpin;
      W &= vpUart_hpin;
      vpUart_hinput = 1;
      if (Zero_) vpUart_hinput = 0;
    }
#endif
  }
#pragma update_FSR 1
}

/**
 * \brief Test if vpUart receiver has a byte waiting
 *
 * This function checks if a byte is available.
 *
 * \n
 * \param W vpUart address.
 * \return True if byte available.
 * \n
 */
bit vpUartRx_byteAvailable(char W) {
#pragma update_FSR 0
  FSR = W;
  if (vpUart_data) return 1;
#pragma update_FSR 1
  return 0;
}

/**
 * \brief Receive byte from vpUart receiver
 *
 * Post-isr code for virtual peripheral receive uart.\n
 * This function grabs a received byte and checks for parity error if required.\n
 *
 * \n
 * How to use:
 * \code
 * if (vpUartRx_byteAvailable(rx)) {
 *    value = vpUartRx_receiveByte(rx);
 * }
 * \endcode
 *
 * \param W vpUart address.
 * \return received byte.
 * \n
 */
char vpUartRx_receiveByte(char W) {
#pragma update_FSR 0
  FSR = W;
  Carry = vpUart_parity; //get last saved b0
  if (vpUart_format.1) goto uit1;
  Carry = 0; //adjust for 7 databits + no parity (total 9 bits)
  vpUart_high = rr(vpUart_high);
  vpUart_high = rr(vpUart_high);
uit1:
  if (vpUart_format.0) vpUart_high = rl(vpUart_high); //adjust for 8 databits + parity (total 11 bits)
  if (!vpUart_7bits) goto uit2;
  Carry = vpUart_high.7; //save possible parity in case 7 databits
  vpUart_high.7 = 0; //clear b7 in case 7 databits
uit2:
#ifdef vpUartRx_parity
  vpUart_parity = Carry; //save possible paritybit
  W = vpUart_high;
  vpUart_low = W;
  //calculate parity
  W = swap(vpUart_low); //_low = abcdefgh
  vpUart_low ^= W;
  W = vpUart_low >> 1;
  vpUart_low ^= W;
  // at this point, the parity for half the bits (a,b,e,f) is in bit 2 of _low, and the
  // parity for the other half (c,d,g,h) is in bit 0 of _low.
  //	if the parity of (a,b,e,f) is 0, then the parity of (a,b,c,d,e,f,g,h)
  // is equal to the parity of (c,d,g,h)... which is already in bit 0, so skip ahead.
  if (vpUart_low.2) {
    // otherwise, the parity of (a,b,e,f) is 1, so the parity of (a,b,c,d,e,f,g,h) is
    // NOT equal to the parity of (c,d,g,h)... invert bit 0.
    vpUart_low++;
  }
  // at this point, bit 0 contains the parity of (a,b,c,d,e,f,g,h).
  vpUart_error = 0;
  W = vpUart_format & 0x18; //extract odd and even bits
  if (!Zero_) {
    if (vpUart_odd) vpUart_low++; //invert calculated paritybit for odd parity
    vpUart_low += vpUart_parity; //if parity matches, b0 will be cleared
    vpUart_error = vpUart_low.0;
  }
#endif
  W = vpUart_high;
  vpUart_data = 0; //mark byte grabbed
#pragma update_FSR 1
  return W;
}

/**
 * \brief Get vpUart receiver error result
 *
 * Post-isr code for virtual peripheral receive uart.\n
 * This function returns the errorbit that is updated by vpUartRx_receiveByte()\n
 * so it should be called after vpUartRx_receiveByte() is called\n
 * (only when parity is used).
 *
 * \n
 * \param W vpUart address.
 * \return errorbit updated by vpUartRx_receiveByte().
 * \n
 */
bit vpUartRx_parityError(char W) {
#pragma update_FSR 0
  FSR = W;
  Carry = vpUart_error; //get error bit
#pragma update_FSR 1
  return Carry;
}

/**
 * \brief Turn vpUart receiver handshake (RTS) on
 *
 * Post-isr code for virtual peripheral receive uart.\n
 * This function turns the optional handshake output pin (RTS) on\n
 * (remote sender may transmit).
 *
 * \n
 * \param W vpUart address.
 * \n
 */
void vpUartRx_hsOn(char W) {
  FSR = W;
#pragma update_FSR 0
  if (vpUart_handshake) {
    vpUart_hlevel = 1;
  }
#pragma update_FSR 1
}

/**
 * \brief Turn vpUart receiver handshake (RTS) off
 *
 * Post-isr code for virtual peripheral receive uart.\n
 * This function turns the optional handshake output pin (RTS) off\n
 * (remote sender may not transmit).
 *
 * \n
 * \param W vpUart address.
 * \n
 */
void vpUartRx_hsOff(char W) {
  FSR = W;
#pragma update_FSR 0
  if (vpUart_handshake) {
    vpUart_hlevel = 0;
  }
#pragma update_FSR 1
}

/**
 * \brief Test if vpUart transmitter is ready
 *
 * Pre-isr code for virtual peripheral transmit uart.\n
 * This part checks if the transmitter is idle and the optional\n
 * handshake input pin (CTS) allows sending.
 *
 * \n
 * \param W vpUart address.
 * \return True if CTS is on or no handshake is used AND the transmitter is idle (transmitting allowed).
 * \n
 */
bit vpUartTx_ready(char W) {
  FSR = W;
#pragma update_FSR 0
  if (vpUart_data) return 0; //transmitter busy
  if (!vpUart_handshake) return 1; //no handshake
  Carry = vpUart_hinput; //current logical CTS input level (1 if transmitting allowed)
#pragma update_FSR 1
  return Carry;
}

/**
 * \brief Write value to vpUart transmitter
 *
 * The function vpUart_sendByte() sends a byte to the vpUart transmitter.\n
 * It must must be tested wether the transmitter is ready to accept a new byte.
 *
 * \n
 * How to use:
 * \code
 * if (vpUartTx_ready(tx)) {
 *   _MainTemp = value; //value to write must be in a global ram location
 *   vpUartTx_sendByte(tx,_MainTemp);
 * }
 * \endcode
 *
 * \param uart vpUart address.
 * \param value Value to write. The value MUST have been stored in a global ram location (0x00-0x0F).
 * \n
 */
#define vpUartTx_sendByte(uart,value) \
{ \
  FSR = uart; \
#pragma update_FSR 0 \
  vpUart_high = ~value; \
#pragma update_FSR 1 \
  _vpUartTx_sendByte(); \
}

/**
 * \brief Send byte to uart transmitter
 *
 * INTERNAL FUNCTION: Pre-isr code for virtual peripheral transmit uart.\n
 * This function calculates parity if required and sets up the transmitter for an outgoing byte.\n
 * Do not call this function directly but call the macro vpUartTx_sendByte().
 */
static void _vpUartTx_sendByte(void) {
#pragma update_FSR 0
  if (vpUart_7bits) vpUart_high.7 = 0; //if not 8 bits then clear b7 for parity calculation (also stopbit)
#ifdef vpUartTx_parity
  //calculate inverted even parity
  W = vpUart_high;
  vpUart_low = W;
  W = swap(vpUart_low); //_low = abcdefgh
  vpUart_low ^= W;
  W = vpUart_low >> 1;
  vpUart_low ^= W;
  // at this point, the even parity for half the bits (a,b,e,f) is in bit 2 of _low, and the
  // even parity for the other half (c,d,g,h) is in bit 0 of _low.
  //	if the even parity of (a,b,e,f) is 0, then the parity of (a,b,c,d,e,f,g,h)
  // is equal to the even parity of (c,d,g,h)... but we want the inverted even parity.
  if (!vpUart_low.2) {
    vpUart_low++;
  }
  vpUart_low = rr(vpUart_low);
  // at this point, carry contains the inverted even parity of (a,b,c,d,e,f,g,h).
  if (vpUart_odd) goto uot1;
  if (!vpUart_even) goto uot2;
    Carry = !Carry; //invert paritybit
uot1:
  if (!vpUart_7bits) goto uot3;
  vpUart_high.7 = Carry; //if 7 databits then move parity into b7
uot2:
#endif
  Carry = 0; //ready stopbit
uot3:
  vpUart_low.7 = 1; //setup startbit
  vpUart_high = rr(vpUart_high); //rotate in paritybit or stopbit
  vpUart_low = rr(vpUart_low);
  if (vpUart_dinvert) {
    vpUart_high = ~vpUart_high; //invert data
    vpUart_low = ~vpUart_low;
  }
  W = vpUart_format; //get total number of bits (start+data+parity+stop)
  W &= 0x03; //1, 2 or 3
  W |= 0x08; //9, 10 or 11
  vpUart_data = 1; //mark transmitter busy
  vpUart_count = W; //now uart vp starts to transmit
#pragma update_FSR 1
}

/**
 * \brief Get address of vpUart queue
 *
 * The function vpUart_queue() returns the address of the char array that is included\n
 * in the vpUart dataset. This array is initialized as a queue, you may however use it as\n
 * you see fit. Just make sure to get its size before using it for some other purpose.\n
 * The reason is that the array has some queue properties that may be overwritten\n
 * when not used as a queue.
 *
 * \n
 * \param W vpUart address.
 * \return Address of queue.
 * \n
 */
char vpUart_queue(char W) {
#pragma update_FSR 0
  W &= 0xF0;
  W |= 0x08; //((&vpUart_hport) & 0x0F);
  FSR = W;
  if (vpUart_handshake) {
    FSR++;
	FSR++;
  }
#pragma update_FSR 1
  return FSR;
}

#pragma library 0
#endif
// End of file vpUart.h
